/*!
 * \file    ds18b20.c
 * \brief   DS18B20 Temperatursensor lesen/schreiben
 * \author  Joerg S.
 * \date    1.2007
 * gendert 04.12.2012 DL9MWE:
   low-level-Routinen: unsigned char OW_Reset (void); unsigned char OW_Read_Bit (void); void OW_Write_Bit (unsigned char bit) 
 */

#include "delay.h"
#include "ds18b20.h"
///////////////////////////////////////////////////////////////////////////////////////////////////
/*!
 * \brief Temperatur DS18B20 messen
 * \param addr  Adresse DS18B20
 * \param temp  Rueckgabepointer fuer Temperatur (Vorkommastellen). MSB ist negativ Bit. Wenn MSb 1,
          dann ist Tempertaur negativ.
 * \param nachkomma Rueckgabepointer fuer Temperatur (Nachkommastellen)
 * \return temp_float Temperatur DS18B20 in float-Variable
 */
float OW_Read_Temp_DS18B20 (unsigned char *ow_addr, unsigned char *temp, unsigned int *nachkomma)
{
  unsigned char get[2];                 // 2 Byte fuer Temperatur Speicher
  float temp_float;                     // 4Byte float
  unsigned char negativ;

  temp_float = 0;
  *temp = 0;
  *nachkomma = 0;

  OW_Read_scratchpad_DS18B20(ow_addr, 2, &get[0]);

  if (get[1] & 0xF0)                    // Wenn oberes Nibble 0xF, dann Temp. negativ
  {
    negativ = TRUE;
    get[1] = ~get[1];                   // Daten negieren
    get[0] = ~get[0];                   // Daten negieren
    ++get[0];                           // Daten +1
  }
  else                                  // Wenn Temp. positiv, dann normal kopieren
  {
    negativ = FALSE;
  }

  temp_float = get[0]>>4 | get[1]<<4;   // Vorkommastellen berechnen
  *temp = (unsigned char) temp_float;   // Zurueckgeben

  if (get[0] & 0x01)                    // Wenn Bit 0 gesetzt, dann...
  {
    temp_float += 0.0625;               // ...0,0625 zur Temperatur addieren
    *nachkomma += 63;
  }
  if (get[0] & 0x02)                    // Wenn Bit 1 gesetzt, dann...
  {
    temp_float += 0.125;                // ...0,125 zur Temperatur addieren
    *nachkomma += 125;
  }
  if (get[0] & 0x04)                    // Wenn Bit 2 gesetzt, dann...
  {
    temp_float +=  0.25;                // ...0,25 zur Temperatur addieren
    *nachkomma += 250;
  }
  if (get[0] & 0x08)                    // Wenn Bit 3 gesetzt, dann...
  {
    temp_float += 0.5;                  // ...0,5 zur Temperatur addieren
    *nachkomma += 500;
  }

  if (negativ) 
  {
  	temp_float *= -1;
  	*temp |= 0x80;   					// Wenn Temp. negativ, dann MSB auf "1"
  }
 
  return temp_float;
}
///////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////
/*!
 * \brief Temperatur DS18B20 messen
 * \param temp  Rueckgabepointer fuer Temperatur (Vorkommastellen)
 * \param nachkomma Rueckgabepointer fuer Temperatur (Nachkommastellen)
 * \param negative True wenn Temperatur negativ, FALSE wenn Temperatur positiv
 * \return temp_float Temperatur DS18B20 in float-Variable
 */
float OW_Read_Temp_DS18B20_s (unsigned char *temp, unsigned int *nachkomma)
{
  unsigned char get[2];                 // 2 Byte fuer Temperatur Speicher
  float temp_float;                     // 4Byte float
  unsigned char negativ;

  temp_float = 0;
  *temp = 0;
  *nachkomma = 0;

  OW_Read_scratchpad_DS18B20_s(2, &get[0]);

  if (get[1] & 0xF0)                    // Wenn oberes Nibble 0xF, dann Temp. negativ
  {
    negativ = TRUE;
    get[1] = ~get[1];                   // Daten negieren
    get[0] = ~get[0];                   // Daten negieren
    ++get[0];                           // Daten +1
  }
  else                                  // Wenn Temp. positiv, dann normal kopieren
  {
    negativ = FALSE;
  }

  temp_float = get[0]>>4 | get[1]<<4;   // Vorkommastellen berechnen
  *temp = (unsigned char) temp_float;   // Zurueckgeben

  if (get[0] & 0x01)                    // Wenn Bit 0 gesetzt, dann...
  {
    temp_float += 0.0625;               // ...0,0625 zur Temperatur addieren
    *nachkomma += 63;
  }
  if (get[0] & 0x02)                    // Wenn Bit 1 gesetzt, dann...
  {
    temp_float += 0.125;                // ...0,125 zur Temperatur addieren
    *nachkomma += 125;
  }
  if (get[0] & 0x04)                    // Wenn Bit 2 gesetzt, dann...
  {
    temp_float +=  0.25;                // ...0,25 zur Temperatur addieren
    *nachkomma += 250;
  }
  if (get[0] & 0x08)                    // Wenn Bit 3 gesetzt, dann...
  {
    temp_float += 0.5;                  // ...0,5 zur Temperatur addieren
    *nachkomma += 500;
  }

  if (negativ) 
  {
  	temp_float *= -1;
  	*temp |= 0x80;   					// Wenn Temp. negativ, dann MSB auf "1"
  }

  return temp_float;
}
///////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////
/*!
 * \brief Seriennummer DS18B20 auslesen
 * \param addr Rueckgabe Pointer fuer 8 Byte Adresse DS18B20
 * \return TRUE wenn Sensor auf Reset geantwortet hat, FALSE bei keiner Reaktion
 * \note  Es darf nur ein Sensor angeschlossen sein!
 */
unsigned char OW_Read_Addr_DS18B20 (unsigned char *ow_addr)
{
  unsigned char answer;
  unsigned int i;

  answer = OW_Reset();                  // One Wire Reset
  OW_Write_Byte (CMD_READ_ROM);         // Read ROM

  for (i = 0; i < 8; i++)               // 8 Byte lesen (1 Byte Family, 6 Byte SNR, 1 Byte CRC)
  {
    ow_addr[i] = OW_Read_Byte();        // Daten einlesen
  }

  return answer;
}
///////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////
/*!
 * \brief Alle DS18B20 Sensoren Temperatur messen
 */
void OW_Convert_Temp_all_DS18B20(void)
{
  OW_Reset();                    // One Wire Reset
  OW_Write_Byte (CMD_SKIP_ROM);  // Skip ROM
  OW_Write_Byte (CMD_CONVERT);   // Convert Temp.
}
///////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////
/*!
 * \brief DS18B20 Sensor Temperatur messen
 * \param addr  Adresse DS18B20
 */
void OW_Convert_Temp_DS18B20(unsigned char *ow_addr)
{
  OW_Reset();                           // One Wire Reset
  OW_Write_Byte (CMD_MATCH_ROM);        // Match ROM (64 bit Adressen benutzen)

  OW_Write_Byte (ow_addr[0]);           // Write Adresse / Family Code
  OW_Write_Byte (ow_addr[1]);           // Write Adresse
  OW_Write_Byte (ow_addr[2]);           // Write Adresse
  OW_Write_Byte (ow_addr[3]);           // Write Adresse
  OW_Write_Byte (ow_addr[4]);           // Write Adresse
  OW_Write_Byte (ow_addr[5]);           // Write Adresse
  OW_Write_Byte (ow_addr[6]);           // Write Adresse
  OW_Write_Byte (ow_addr[7]);           // Write Adresse / CRC

  OW_Write_Byte (CMD_CONVERT);          // Convert Temp.
}
///////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////
/*!
 * \brief DS18B20 Scratchpad Register beschreiben
 * \param addr  Adresse DS18B20
 * \param data_user_1 User-Byte 1
 * \param data_user_2 User-Byte 2
 * \param data_config Wert der in Configuration-Register geschrieben werden soll
 */
void OW_Write_config_DS18B20 (unsigned char *ow_addr,
                              unsigned char data_user_1,
                              unsigned char data_user_2,
                              unsigned char data_config)
{
  OW_Reset();                           // One Wire Reset
  OW_Write_Byte (CMD_MATCH_ROM);        // Match ROM (64 bit Adressen benutzen)

  OW_Write_Byte (ow_addr[0]);           // Write Adresse / Family Code
  OW_Write_Byte (ow_addr[1]);           // Write Adresse
  OW_Write_Byte (ow_addr[2]);           // Write Adresse
  OW_Write_Byte (ow_addr[3]);           // Write Adresse
  OW_Write_Byte (ow_addr[4]);           // Write Adresse
  OW_Write_Byte (ow_addr[5]);           // Write Adresse
  OW_Write_Byte (ow_addr[6]);           // Write Adresse
  OW_Write_Byte (ow_addr[7]);           // Write Adresse / CRC

  OW_Write_Byte (CMD_WR_SCRATCHPAD);    // Write Scartchpad
  OW_Write_Byte (data_user_1);          // Schreibe TH Register
  OW_Write_Byte (data_user_2);          // Schreibe TL Register
  OW_Write_Byte (data_config);          // Schreibe Configuration Register
}
///////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////
/*!
 * \brief DS18B20 Scratchpad Register beschreiben
 * \param data_user_1 User-Byte 1
 * \param data_user_2 User-Byte 2
 * \param data_config Wert der in Configuration-Register geschrieben werden soll
 */
void OW_Write_config_DS18B20_s (unsigned char data_user_1,
                                unsigned char data_user_2,
                                unsigned char data_config)
{
  OW_Reset();                           // One Wire Reset
  OW_Write_Byte (CMD_SKIP_ROM);         // Skip ROM (Keine Adressen benutzen)

  OW_Write_Byte (CMD_WR_SCRATCHPAD);    // Write Scartchpad
  OW_Write_Byte (data_user_1);          // Schreibe TH Register
  OW_Write_Byte (data_user_2);          // Schreibe TL Register
  OW_Write_Byte (data_config);          // Schreibe Configuration Register
}
///////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////
/*!
 * \brief DS18B20 Scratchpad lesen
 * \param addr Adresse DS18B20
 * \param cnt  Anzahl der Bytes die gelesen werden sollen (maximal 9)
 * \param data Rueckgabepointer fuer Daten-Array
 */
void OW_Read_scratchpad_DS18B20 (unsigned char *ow_addr, unsigned char cnt, unsigned char *data)
{
  unsigned char i;

  OW_Reset();                           // One Wire Reset
  OW_Write_Byte (CMD_MATCH_ROM);        // Match ROM (64 bit Adressen benutzen)

  OW_Write_Byte (ow_addr[0]);           // Write Adresse / Family Code
  OW_Write_Byte (ow_addr[1]);           // Write Adresse
  OW_Write_Byte (ow_addr[2]);           // Write Adresse
  OW_Write_Byte (ow_addr[3]);           // Write Adresse
  OW_Write_Byte (ow_addr[4]);           // Write Adresse
  OW_Write_Byte (ow_addr[5]);           // Write Adresse
  OW_Write_Byte (ow_addr[6]);           // Write Adresse
  OW_Write_Byte (ow_addr[7]);           // Write Adresse / CRC

  OW_Write_Byte (CMD_RD_SCRATCHPAD);    // Read Scartchpad

  for (i = 0; i < cnt; i++)             // Scratchpad lesen, maximal 9 Byte
  {
    data[i] = OW_Read_Byte();           // Daten einlesen
  }

  OW_Reset();                           // One Wire Reset
}
///////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////
/*!
 * \brief DS18B20 Scratchpad lesen
 * \param cnt  Anzahl der Bytes die gelesen werden sollen (maximal 9)
 * \param data Rueckgabepointer fuer Daten-Array
 */
void OW_Read_scratchpad_DS18B20_s (unsigned char cnt, unsigned char *data)
{
  unsigned char i;

  OW_Reset();                           // One Wire Reset
  OW_Write_Byte (CMD_SKIP_ROM);         // Skip ROM (Keine Adressen benutzen)

  OW_Write_Byte (CMD_RD_SCRATCHPAD);    // Read Scartchpad

  for (i = 0; i < cnt; i++)             // Scratchpad lesen, maximal 9 Byte
  {
    data[i] = OW_Read_Byte();           // Daten einlesen
  }

  OW_Reset();                           // One Wire Reset
}
///////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////
/*!
 * \brief DS18B20 Scratchpad in EEPROM speichern
 * \param addr Adresse DS18B20
 */
void OW_Copy_scratchpad_DS18B20 (unsigned char *ow_addr)
{
  OW_Reset();                           // One Wire Reset
  OW_Write_Byte (CMD_MATCH_ROM);        // Match ROM (64 bit Adressen benutzen)

  OW_Write_Byte (ow_addr[0]);           // Write Adresse / Family Code
  OW_Write_Byte (ow_addr[1]);           // Write Adresse
  OW_Write_Byte (ow_addr[2]);           // Write Adresse
  OW_Write_Byte (ow_addr[3]);           // Write Adresse
  OW_Write_Byte (ow_addr[4]);           // Write Adresse
  OW_Write_Byte (ow_addr[5]);           // Write Adresse
  OW_Write_Byte (ow_addr[6]);           // Write Adresse
  OW_Write_Byte (ow_addr[7]);           // Write Adresse / CRC

  OW_Write_Byte (CMD_CP_SCRATCHPAD);    // Copy Scartchpad

//  delay_ms(10);                         // Speicher schreiben benoetigt max. 10ms
}
///////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////
/*!
 * \brief DS18B20 Scratchpad in EEPROM speichern
 */
void OW_Copy_scratchpad_DS18B20_s (void)
{
  OW_Reset();                           // One Wire Reset
  OW_Write_Byte (CMD_SKIP_ROM);         // Skip ROM (Keine Adressen benutzen)

  OW_Write_Byte (CMD_CP_SCRATCHPAD);    // Copy Scartchpad

//  delay_ms(10);                         // Speicher schreiben benoetigt max. 10ms
}
///////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////
/*!
 * \brief Reset fuer OneWire
 * \return answer TRUE wenn Endgeraet geantwortet hat, FALSE wenn nicht
 */
unsigned char OW_Reset (void)
{
  unsigned char answer;

	P1DIR |= DS18B20PIN;		// DS18B20-Pin ist Ausgang
 	P1OUT &= ~DS18B20PIN;		// Reset-Pulse (low fr >480us)
 	_delay_us(500);
 	P1DIR &= ~DS18B20PIN;		// 1-Wire als Eingang schalten und via Pull-Up auf 1 ziehen
 	_delay_us(75);				//  und dem DS18B20 >60us Zeit lassen
	if (DS18B20_IN_H == FALSE)
		answer = TRUE; 				// OnwWire ist low (Antwort von Geraet)
	else 
		answer = FALSE;             // OneWire ist high (Keine Antwort)
  _delay_us(500);
  return answer;
}
///////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////
/*!
 * \brief Bit lesen auf OneWire
 * \return result Gelesenes Bit von OneWire
 */
unsigned char OW_Read_Bit (void)
{
  unsigned char result;
	P1DIR |= DS18B20PIN;		// DS18B20-Pin ist Ausgang
 	P1OUT &= ~DS18B20PIN;		// Low-Puls (fr >= 1uS)
 	_delay_us(2);
 	P1DIR &= ~DS18B20PIN;		// 1-Wire als Eingang schalten und dem DS18B20 berlassen
 	_delay_us(10);				// (Ergebnis muss innerhalb von 15uS gelesen werden)
	result = DS18B20_IN_H;		// Pegel von Datenleitung = Ergebnis
 	_delay_us(60);				// Gesamtzeit >= 60uS	
  return result;                // Wert zurueckgeben
}
///////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////
/*!
 * \brief Byte lesen auf OneWire
 * \return result Gelesenes Byte von OneWire
 */
unsigned char OW_Read_Byte (void)
{
  int loop, result = 0;

  for (loop = 0; loop < 8; loop++)
  {
    result >>= 1;               			// Shift fuer naechstes Bit
    if (OW_Read_Bit()) result |= 0x80;    	// Wenn gelesenes Bit = 1, dann Bit setzen
  }

  return result;
}
///////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////
/*!
 * \brief Bit schreiben auf OneWire
 * \param bit Zu schreibendes Bit auf OneWire
 */
void OW_Write_Bit (unsigned char bit)
{
  if (bit)                      // Wenn bit = 1 ...
  {
	P1DIR |= DS18B20PIN;		// DS18B20-Pin ist Ausgang
 	P1OUT &= ~DS18B20PIN;		// Low-Puls (fr < 15uS)
 	_delay_us(13);				//  (passt zu H13)
 	P1DIR &= ~DS18B20PIN;		// 1-Wire als Eingang schalten und via Pull-Up auf 1 ziehen
 	_delay_us(47);				//  und dem DS18B20 Zeit lassen, die 1 zu lesen	
  }
  else                          // Wenn bit = 0 ...
  {
	P1DIR |= DS18B20PIN;		// DS18B20-Pin ist Ausgang
 	P1OUT &= ~DS18B20PIN;		// Low-Puls (fr >60uS)
 	_delay_us(75);				// DS18B20 liest die 0
 	P1DIR &= ~DS18B20PIN;		// 1-Wire als Eingang schalten und via Pull-Up auf 1 ziehen
  }
  _delay_us(1);
}
///////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////
/*!
 * \brief Byte schreiben auf OneWire
 * \param data Zu schreibendes Byte auf OneWire
 */
void OW_Write_Byte (unsigned char data)
{
  int loop;

  for (loop = 0; loop < 8; loop++)
  {
    OW_Write_Bit (data & 0x01);
    data >>= 1;                 // Shift data fuer naechstes Bit
  }
}
///////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////
